/*
 * Freshmeat retrieval tool.
 *
 * (C) 1998 Alan Cox.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */
 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h> 
#include <errno.h>
#include <sys/time.h>

#include "config.h"
#include "portaloo.h"

static int host_write(int fd, char *ptr)
{
	int err=write(fd, ptr, strlen(ptr));
	return err;
}

static char host_rbuf[4096];
static int host_rptr;
static int host_rlen;

static int host_getblk(int fd)
{
	int rlen;
	fd_set rfd;
	struct timeval tv;
	
	while(1)
	{
		rlen=read(fd, host_rbuf, sizeof(host_rbuf));
		if(rlen>0)
		{
			host_rlen=rlen;
			host_rptr=0;
			return host_rlen;
		}
		if(rlen==0)
		{
			return 0;
		}
		if(rlen==-1 && errno!=EWOULDBLOCK)
		{
			return 0;
		}
		tv.tv_sec=60;
		tv.tv_usec=0;
		FD_ZERO(&rfd);
		FD_SET(fd, &rfd);
		
		if(select(fd+1, &rfd, NULL, NULL, &tv)<1)
			return 0;
	}
}

char *host_readline(int fd)
{
	static char buf[4096];
	char *bp=buf;
	
	while(bp-buf<4095)
	{
		if(host_rptr==host_rlen)
			if(host_getblk(fd)==0)
				return NULL;
		*bp=host_rbuf[host_rptr++];
		if(*bp=='\n')
		{
			*bp=0;
			printf(">%s\n",buf);
			return buf;
		}
		if(*bp!='\r')
			bp++;
	}
}

	

static int host_connect_attempt(struct in_addr ia, int port)
{
	int s=socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	
	struct sockaddr_in sin;
	
	fd_set wfd;
	struct timeval tv;
	
	if(s==-1)
	{
		perror("socket");
		return -1;
	}
	
	if(fcntl(s, F_SETFL, FNDELAY)==-1)
	{
		perror("nonblocking");
		close(s);
		return -1;
	}

	sin.sin_family = AF_INET;	
	sin.sin_addr   = ia;
	sin.sin_port   = htons(port);
	
	if(connect(s, (struct sockaddr *)&sin, sizeof(sin))==-1 && errno != EINPROGRESS)
	{
		perror("connect");
		close(s);
		return -1;
	}	
	
	tv.tv_sec = 60;		/* We use 60 second timeouts for now */
	tv.tv_usec = 0;
	
	FD_ZERO(&wfd);
	FD_SET(s, &wfd);
	
	switch(select(s+1, NULL, &wfd, NULL, &tv))
	{
		case 0:
			/* Time out */
			close(s);
			return -1;
		case -1:
			/* Ermm.. ?? */
			perror("select");
			close(s);
			return -1;
	}
	
	return s;
}
 
int host_connect(const char *host, int port)
{
	struct hostent *h;
	int i;
	int s;
	
	h=gethostbyname(host);
	if(h==NULL)
	{
		fprintf(stderr,"unable to resolve '%s'.\n", host);
		return -1;
	}
	
	
	for(i=0; h->h_addr_list[i]; i++)
	{
		struct in_addr ia;
		memcpy(&ia, h->h_addr_list[i],4);
		s=host_connect_attempt(ia, port);
		if(s != -1)
			return s;
	}
	fprintf(stderr, "unable to connect to '%s'.\n", host);
	return -1;
}

static void date_smash(char *p, int *mm, int *dd, int *h, int *m)
{
	char ap;
	
	*dd = -1;
	*mm = -1;

#if 0
	switch(*p)
	{
		case 'M':
			*dd=1;
			break;
		case 'T':
			*dd=2;
			if(p[1]=='h')
				*dd=4;
			break;
		case 'W':
			*dd=3;
			break;
		case 'F':
			*dd=5;
			break;
		case 'S':
			*dd=6;
			if(p[1]=='u')
				*dd=7;
	}
	while(*p && !isspace(*p))
		p++;
#endif
		
	while(*p && isspace(*p))
		p++;
	switch(*p)
	{
		case 'J':
			*mm=1;
			if(p[1]=='u')
				*mm=6;
			if(p[2]=='l')
				*mm=7;
			break;
		case 'F':
			*mm=2;
			break;
		case 'M':
			*mm=3;
			if(p[2]=='y')
				*mm=5;
			break;
		case 'A':
			*mm=4;
			if(p[1]=='u')
				*mm=8;
			break;
		case 'S':
			*mm=9;
			break;
		case 'O':
			*mm=10;
			break;
		case 'N':
			*mm=11;
			break;
		case 'D':
			*mm=12;
			break;
	}
	while(*p && !isspace(*p))
		p++;
	sscanf(p, " %d, %d:%d %c", dd, h, m, &ap);
	if(ap=='p')
		*h+=12;
}
	
int main(int argc, char *argv[])
{
	int s;
	int head=1;
	int ct=0;	
	char *p;
	char buf[256];
	char base[256];
	FILE *f;
	
	if(argc!=2)
	{
		fprintf(stderr,"%s [num]\n", argv[0]);
		exit(1);
	}	
	snprintf(buf, sizeof(buf), PORTALOO_BASE"%s.new", argv[1]);
	snprintf(base, sizeof(buf), PORTALOO_BASE"%s.out", argv[1]);
	
	f=fopen(buf, "w");
	if(f==NULL)
	{
		perror(buf);
		exit(1);
	}
	
	s=host_connect("files.freshmeat.net",80);
	if(s==-1)
		return 1;
	
	host_write(s,"GET /freshmeat/recentnews.txt HTTP/1.0\r\nhost: files.freshmeat.net\r\n\r\n");
	
	fprintf(f,"Freshmeat.Net\nhttp://freshmeat.net\nFreshmeat: Software Index\n--\n");
	fprintf(f,"<br clear=all><FONT size=\"-1\" face=\"helvetica\"><P><UL>\n");
	
	while((p=host_readline(s))!=NULL)
	{
		char *app;
		if(head && *p!=0 && *p!='\r')
			continue;
		
		if(head)
		{		
			head = 0;
			continue;
		}
	
		app=strdup(p);
		
		if((p=host_readline(s))==NULL)
			break;
	
		if((p=host_readline(s))==NULL)
			break;
		/*
		 *	Ok its now loaded a block for one app. We need
		 *	to process this up
		 */
		
		fprintf(f,"<LI><A href=\"%s\">%s</A>\n",
			p, app);
		ct++;
		if(ct==10)
			fprintf(f,"--\n");
		free(app);
	}
	fprintf(f,"</UL></FONT>\n");
	fclose(f);
	if(rename(buf,base)==-1)
	{
		perror("rename");
		exit(1);
	}
	exit(0);
}
	