/*
 * Quote retrieval tool.
 *
 * (C) 1998 Alan Cox.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */
 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h> 
#include <errno.h>
#include <sys/time.h>

#include "config.h"
#include "portaloo.h"

static int host_write(int fd, char *ptr)
{
	int err=write(fd, ptr, strlen(ptr));
	return err;
}

static char host_rbuf[4096];
static int host_rptr;
static int host_rlen;

static int host_getblk(int fd)
{
	int rlen;
	fd_set rfd;
	struct timeval tv;
	
	while(1)
	{
		rlen=read(fd, host_rbuf, sizeof(host_rbuf));
		if(rlen>0)
		{
			host_rlen=rlen;
			host_rptr=0;
			return host_rlen;
		}
		if(rlen==0)
		{
			return 0;
		}
		if(rlen==-1 && errno!=EWOULDBLOCK)
		{
			return 0;
		}
		tv.tv_sec=60;
		tv.tv_usec=0;
		FD_ZERO(&rfd);
		FD_SET(fd, &rfd);
		
		if(select(fd+1, &rfd, NULL, NULL, &tv)<1)
			return 0;
	}
}

char *host_readline(int fd)
{
	static char buf[4096];
	char *bp=buf;
	
	while(bp-buf<4095)
	{
		if(host_rptr==host_rlen)
			if(host_getblk(fd)==0)
				return NULL;
		*bp=host_rbuf[host_rptr++];
		if(*bp=='\n')
		{
			*bp=0;
			printf(">%s\n",buf);
			return buf;
		}
		if(*bp!='\r')
			bp++;
	}
}

	

static int host_connect_attempt(struct in_addr ia, int port)
{
	int s=socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	
	struct sockaddr_in sin;
	
	fd_set wfd;
	struct timeval tv;
	
	if(s==-1)
	{
		perror("socket");
		return -1;
	}
	
	if(fcntl(s, F_SETFL, FNDELAY)==-1)
	{
		perror("nonblocking");
		close(s);
		return -1;
	}

	sin.sin_family = AF_INET;	
	sin.sin_addr   = ia;
	sin.sin_port   = htons(port);
	
	if(connect(s, (struct sockaddr *)&sin, sizeof(sin))==-1 && errno != EINPROGRESS)
	{
		perror("connect");
		close(s);
		return -1;
	}	
	
	tv.tv_sec = 60;		/* We use 60 second timeouts for now */
	tv.tv_usec = 0;
	
	FD_ZERO(&wfd);
	FD_SET(s, &wfd);
	
	switch(select(s+1, NULL, &wfd, NULL, &tv))
	{
		case 0:
			/* Time out */
			close(s);
			return -1;
		case -1:
			/* Ermm.. ?? */
			perror("select");
			close(s);
			return -1;
	}
	
	return s;
}
 
int host_connect(const char *host, int port)
{
	struct hostent *h;
	int i;
	int s;
	
	h=gethostbyname(host);
	if(h==NULL)
	{
		fprintf(stderr,"unable to resolve '%s'.\n", host);
		return -1;
	}
	
	
	for(i=0; h->h_addr_list[i]; i++)
	{
		struct in_addr ia;
		memcpy(&ia, h->h_addr_list[i],4);
		s=host_connect_attempt(ia, port);
		if(s != -1)
			return s;
	}
	fprintf(stderr, "unable to connect to '%s'.\n", host);
	return -1;
}

int main(int argc, char *argv[])
{
	int s;
	int head=1;
	int ct=0;	
	char *p;
	char buf[256];
	char base[256];
	FILE *f;
	char *stock_code;
	char *stock_name;
	
	if(argc!=4)
	{
		fprintf(stderr,"%s [num] [code] [name]\n", argv[0]);
		exit(1);
	}	
	
	stock_code=argv[2];
	stock_name=argv[3];
	
	snprintf(buf, sizeof(buf), PORTALOO_BASE"%s.new", argv[1]);
	snprintf(base, sizeof(buf), PORTALOO_BASE"%s.out", argv[1]);
	
	f=fopen(buf, "w");
	if(f==NULL)
	{
		perror(buf);
		exit(1);
	}
	
	s=host_connect("quote.yahoo.com",80);
	if(s==-1)
		return 1;
	
	host_write(s,"GET /d/quotes.csv?f=sl1d1t1c1ohgv&e=.csv&s=");
	host_write(s, stock_code);
	host_write(s," HTTP/1.0\r\nhost: quote.yahoo.com\r\n\r\n");
	
	fprintf(f,"%s Stock\nhttp://quote.yahoo.com\nYahoo: %s Stock\n--\n", stock_name, stock_name);
	fprintf(f,"<br clear=all><FONT size=\"-1\" face=\"helvetica\"><P>\n");
	
	while((p=host_readline(s))!=NULL)
	{
		char *app;
		char *prc;
		char *time;
		char *shift;
		if(head && *p!=0 && *p!='\r')
			continue;
		
		if(head)
		{		
			head = 0;
			continue;
		}
	
		p++;	
		app=strchr(p,'"');
		if(app==NULL)
			continue;
		*app=0;
		app+=2;
		prc=app;
		app=strchr(app, ',');
		if(app==NULL)
			continue;
		*app++=0;
		app=strchr(app, ',');
		if(app==NULL)
			continue;
		/* Ok we are now at the time */
		app++;
		if(*app!='"')
			continue;
		app++;	/* " */	
		time=app;
		app=strchr(app,'"');
		if(app==NULL)
			continue;
		*app=0;
		app+=2;
		shift=app;
		
		app=strchr(app, ',');
		if(app==NULL)
			continue;
		*app=0;
		
		fprintf(f,"<CENTER><FONT size=+2>%s</FONT> <FONT size=-1>at %s</FONT><br><br>\n",
			stock_code, time);
		fprintf(f,"<br><br><br><FONT size=+2>%s&nbsp;", prc);
		if(*shift=='-')
			fprintf(f,"<FONT color=red>");
		else if(*shift=='+')
			fprintf(f,"<FONT color=green>");
		else fprintf(f,"<FONT color=blue>");
		fprintf(f,"%s</FONT></FONT></CENTER><br>\n", shift);		
	}
	fprintf(f,"</FONT>\n");
	fclose(f);
	if(rename(buf,base)==-1)
	{
		perror("rename");
		exit(1);
	}
	exit(0);
}
	