#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "config.h"
#include "struct.h"
#include "thread.h" 
#include "proto.h"

struct thread_tree  *articles;
char *thread_folder=NULL;
int thread_size=0;

int load_thread(char *name)
{
	int file;
	struct Header ff;
	char path[512];
	int first, last, size;
	int i;
	struct thread_tree *ptr;

	snprintf(path, 512, "%s/%s.i", HOMEPATH, name);
	if ((file=open(path,O_RDONLY))<0)
		html_error(name);

	thread_folder=strdup(name);

	/* determine the size of the tree by seeking to the ends */
	if (read(file, &ff, sizeof(ff))<sizeof(ff))
	{
		thread_size=0;
		return(0);
	}
	first=ff.Ref;
	lseek(file, -1 * sizeof(ff), SEEK_END);
	read(file, &ff, sizeof(ff));
	last=ff.Ref;
	size=(last-first)+1;
	thread_size=size;

	articles=(struct thread_tree *)calloc(size, sizeof(struct thread_tree));

#ifdef DEBUG
	printf("Loading articles %d to %d = %d: ", first, last, size);
#endif	
	/* plough through and load the tree */	
	lseek(file, 0, SEEK_SET);
	while (read(file, &ff, sizeof(ff))>0)
	{
		i=ff.Ref-first;
		articles[i].ref=-1;
		articles[i].depth=0;
		articles[i].parent=NULL;
		articles[i].child=NULL;
		articles[i].next=NULL;		
		memcpy(&(articles[i].article), &ff, sizeof(ff));
#ifdef DEBUG
		printf("%d ", i);
#endif
	}

	articles[0].ref=0;
#ifdef DEBUG
		printf("\n\nStarting threading...\n");
#endif
	for (i=1; i<size; i++)
	{
		int j;
		if (articles[i].ref != -1) continue;

		articles[i].ref=articles[i].article.Ref;
		/* article that this is a reply to */
		j=articles[i].article.replyto;
		ptr=&articles[j-first];
#ifdef DEBUG
		printf("%d is reply to %d (got %d)", articles[i].ref,
			j, ptr->ref);
#endif
		if (ptr->child==NULL)
		{
#ifdef DEBUG
printf(", first child.\n");
#endif 
			ptr->child=&articles[i];
			articles[i].parent=ptr;
			articles[i].next=NULL;
			articles[i].depth=ptr->depth+1;
		}else
		{
			articles[i].parent=ptr;
			articles[i].next=NULL;
			articles[i].depth=ptr->depth+1;
			ptr=ptr->child;
			while (ptr->next!=NULL)
				ptr=ptr->next;
			ptr->next=&articles[i];
#ifdef DEBUG
printf(", later child, after %d\n", ptr->ref);
#endif
		}
	}
	return (size);
}


/*** display structure ***/

int thread_count( struct thread_tree *ptr )
{
	int count=0;

	while (ptr!=NULL)
	{
		if (ptr->child!=NULL)
			count += thread_count(ptr->child);
		count++;
		ptr=ptr->next;
	}
	return (count);
}
	
static void depth_print( int depth , unsigned long mask)
{
	int i;
	for (i=0;i<depth; i++)
	if ( mask&(1<<i))
		printf("<img src=%s/tree/blank4.gif alt=\" \" align=top>", IMAGEDIR);
	else
		printf("<img src=%s/tree/blank1.gif alt=| align=top>", IMAGEDIR);
}

void article_short(struct Header article, int depth, int maxdepth)
{
	if (article.icon>0) printf(" %s", small_icon(article.icon));

	printf("  <a href=%s?fol=%s&msg=%d>%s</a> from %s ",
		CGIPATH, thread_folder, article.Ref, article.subject, article.from);
//	printf(" <a href=%s?fol=%s&replyto=%d&action=comment>Reply</a>", CGIPATH, thread_folder, article.Ref);
	printf("</td></tr>\n");
}

void article_full(struct Header article, int depth, int maxdepth, int hdr)
{
	char *text;
	int ff;
	char path[513];

	snprintf(path, 512, "%s/%s.t", HOMEPATH, thread_folder);
	if ((ff=open(path, O_RDONLY))<0)
		text=strdup("No Text");
	else
	{
		int size=0;
		lseek(ff, article.datafield, SEEK_SET);
		size=article.size;
		text=(char *)malloc(size +1);
		read(ff, text, size);
		text[size]=0;
		close(ff);
	}
	
	if(hdr)
	{
		printf("%s\n", article.subject);
		return;
	}

	if (article.Ref==0)
	{
		printf("<table border=0 width=98%% cellspacing=0>\n");
		printf("<tr><td bgcolor=#000000><FONT size=+1 color=white face=helvetica><b>%s</b></FONT></td></tr>\n", 
			article.subject);
		printf("<tr><td bgcolor=#eeeebb>");
		printf("<font face=helvetica><b>Posted by %s on %s</b></font><br>\n",
			article.from, ctime(&article.date) );
		printf("<p><FONT face=helvetica>%s</FONT></td></tr>\n", text);
	}
	else
	{
		printf("<table border=0 width=98%% cellspacing=0>\n");
		printf("<tr><td bgcolor=#ddddaa>");
		printf("<Font face=helvetica>");
		printf("<b>");
		if (article.icon>0) printf("%s ",small_icon(article.icon));
		printf("%s</b><br>\n", article.subject);
		printf("Comment %d posted by %s on %s</FONT>",
			article.Ref, article.from, ctime(&article.date) );
		printf("</td></tr>\n");

		printf("<tr><td bgcolor=#eeeebb><FONT face=helvetica>%s</FONT></td></tr>\n", text);
	}

	if (depth+1< maxdepth)
		printf("<tr><td bgcolor=#eeeebb align=right>\n");
	else
		printf("<tr><td bgcolor=#eeeebb align=left>\n");

	free(text);
}

static int wasfull=1;

static void thread_recurse(struct thread_tree *ptr, int depth, int maxdepth, unsigned long mask)
{
	int topen=0;
	int treeopen=0;

	while (ptr!=NULL)
	{
		if (depth>=maxdepth)
		{
			if (wasfull) 
			{
				printf("<br><br><table frame=void cellspacing=0 cellpadding=0 border=0>\n");
				printf("<tr><td nowrap><img src=%s/tree/open_folder.gif align=top alt=o> %d comments.</td></tr>\n", IMAGEDIR, thread_count( ptr ));
				treeopen=1;
			}

			printf("<tr><td nowrap>");
			depth_print(depth-maxdepth, mask);
			if (ptr->next==NULL)
			{
				printf("<img src=%s/tree/blank3.gif align=top alt=+>", IMAGEDIR);
				mask|=1<<(depth-maxdepth);
			}else
				printf("<img src=%s/tree/blank2.gif align=top alt=+>", IMAGEDIR);
			if (ptr->child==NULL)
				printf("<img src=%s/tree/book.gif align=top alt=o>", IMAGEDIR);
			else
				printf("<img src=%s/tree/open_book.gif align=top alt=o>", IMAGEDIR);
			article_short( ptr->article, depth , maxdepth);
			wasfull=0;
		}
		else
		{
			article_full( ptr->article, depth, maxdepth , 0);
			printf("<FONT face=helvetica size=-1><a href=%s?fol=%s&replyto=%d&action=comment>Add Comment</a></FONT>\n", CGIPATH, thread_folder, ptr->ref);
			wasfull=1;
			topen=1;
		}
		if ( ptr->child !=NULL )
			thread_recurse( ptr->child, depth+1, maxdepth, mask );
		if (topen) printf("<br></td></tr></table>\n");
		ptr=ptr->next;
	}
	if (treeopen) printf("</table>\n");
}

void show_thread(int size, int maxdepth, int id)
{
	int depth=0;
	int i;

	print_header("Portaloo - Full Article");
	i=id;

	if (i>=size)
	{
		printf("No Article found.\n");
	}else
	{
		depth=articles[i].depth;
		article_full(articles[i].article, depth, maxdepth, 0); 

		if (articles[i].child !=NULL) {
		/* preferences menu block */
			printf("\n<font size=-3>");
			printf("<form action=%s method=get>\n", CGIPATH);
			printf("<input type=hidden name=cookie_pref value=depth>\n");
			printf("<input type=hidden name=fol value=%s>\n", thread_folder);
			if (id>0) 
				printf("<input type=hidden name=msg value=%d>\n", id);
			printf("<table border=0 cellspacing=0 cellpadding=0 bgcolor=\"#eeeebb\" width=99%%><tr>\n");
			printf("<td><a href=%s?fol=%s&replyto=%d&action=comment>Add Comment</a>&nbsp;&nbsp;&nbsp;</td>", CGIPATH, thread_folder, articles[i].ref);
			if(articles[i].parent)
				printf("<td>&nbsp;<a href=%s?fol=%s&msg=%d>Parent</a></td>", CGIPATH, thread_folder, articles[articles[i].parent->article.Ref].ref);
			else
				printf("<td>&nbsp;<a href=%s>Articles</A>&nbsp;&nbsp;</td>", CGIPATH);
			printf("<td><FONT face=helvetica>Depth ");
			printf("<input type=radio name=depth value=0 %s>None\n", maxdepth==0?"checked":"");
			printf("<input type=radio name=depth value=1 %s>1\n", maxdepth==1?"checked":"");
			printf("<input type=radio name=depth value=2 %s>2\n", maxdepth==2?"checked":"");
			printf("<input type=radio name=depth value=3 %s>3\n", maxdepth==3?"checked":"");
			printf("<input type=radio name=depth value=255 %s>All\n", maxdepth>4?"checked":"");
			printf(" <input type=submit name=submit value=Set>\n");
			printf("</font></td></tr></table></form></font>\n");
		}
		else
			printf("<br><a href=%s?fol=%s&replyto=%d&action=comment>Add Comment</a>\n", CGIPATH, thread_folder, articles[i].ref);			


		if (articles[i].child !=NULL)
			thread_recurse( articles[i].child , 0, maxdepth, 0 );
		printf("</td></tr></table>\n");
	}
	print_footer();
}

