#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <signal.h>

#include <netinet/in.h>
#include <sys/socket.h>

#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <sys/time.h>

#include <sys/un.h>

#include "config.h"
#include "cgi.h"
#include "portaloo.h"
#include "md5.h"
#include "web_tty.h"

#define LINE_ROWS	38
#define LINE_KEEP	20
#define LINE_COLS	80

static char data[LINE_ROWS][LINE_COLS];
static int xp, yp;
static int redraw = 1;

static int host_connect_attempt(struct in_addr ia, int port)
{
	int s=socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	
	struct sockaddr_in sin;
	
	fd_set wfd;
	struct timeval tv;
	
	if(s==-1)
	{
		perror("socket");
		return -1;
	}
	
	if(fcntl(s, F_SETFL, FNDELAY)==-1)
	{
		perror("nonblocking");
		close(s);
		return -1;
	}

	sin.sin_family = AF_INET;	
	sin.sin_addr   = ia;
	sin.sin_port   = htons(port);
	
	if(connect(s, (struct sockaddr *)&sin, sizeof(sin))==-1 && errno != EINPROGRESS)
	{
		perror("connect");
		close(s);
		return -1;
	}	
	
	tv.tv_sec = 60;		/* We use 60 second timeouts for now */
	tv.tv_usec = 0;
	
	FD_ZERO(&wfd);
	FD_SET(s, &wfd);
	
	switch(select(s+1, NULL, &wfd, NULL, &tv))
	{
		case 0:
			/* Time out */
			close(s);
			return -1;
		case -1:
			/* Ermm.. ?? */
			perror("select");
			close(s);
			return -1;
	}
	
	return s;
}

int host_connect(const char *host, int port)
{
	struct hostent *h;
	int i;
	int s;
	
	h=gethostbyname(host);
	if(h==NULL)
	{
		fprintf(stderr,"unable to resolve '%s'.\n", host);
		return -1;
	}
	
	
	for(i=0; h->h_addr_list[i]; i++)
	{
		struct in_addr ia;
		memcpy(&ia, h->h_addr_list[i],4);
		s=host_connect_attempt(ia, port);
		if(s != -1)
			return s;
	}
	fprintf(stderr, "unable to connect to '%s'.\n", host);
	return -1;
}
	
static void scroll_check(void)
{
	int i;
	if(yp<LINE_ROWS)
		return;
	for(i=0;i<LINE_KEEP;i++)
		memcpy(data[i], data[i+LINE_ROWS-LINE_KEEP], LINE_COLS);
	for(i=LINE_KEEP; i< LINE_ROWS; i++)
		memset(data[i], ' ', LINE_COLS);
	yp=LINE_KEEP;
	redraw=1;
}
	

static void print_string(char *s)
{
	while(*s)
	{
		if(*s=='\r')
			xp=0;
		else if(*s=='\n')
		{
			xp=0;
			yp++;
			scroll_check();
		}
		else if(*s=='\t')
		{
			do
			{
				print_string(" ");
			}
			while(xp%8);
		}
		else
		{
			if(xp==LINE_COLS)
			{
				xp=0;
				yp++;
				scroll_check();
			}
			data[yp][xp++]=*s;
		}
		s++;
	}
}

static char *trimify(char *p)
{
	char buf[LINE_COLS+1];
	static char workbuf[512];
	int i;
	char *bp;
	char *op;
	
	memcpy(buf,p,LINE_COLS);
	buf[LINE_COLS]=0;
	
	for(i=LINE_COLS-1;i>0;i--)
	{
		if(buf[i]!=' ')
			break;
		buf[i]=0;
	}
	bp=buf;
	op=workbuf;
	while(*bp)
	{
		if(*bp=='<')
		{
			strcpy(op, "&lt;");
			op+=4;
		}
		else if(*bp=='>')
		{
			strcpy(op, "&gt;");
			op+=4;
		}
		else if(*bp==' ')
		{
			strcpy(op, "&nbsp;");
			op+=6;
		}
		else if(*bp=='&')
		{
			strcpy(op, "&amp;");
			op+=5;
		}
		else *op++=*bp;
		bp++;
	}
	*op=0;
	return workbuf;
}

void weboid_sync(void)
{
	int i;
	static int lyp = 0;
	
	if(redraw)
	{
		printf("\n--magicmetamarkerthing--\n");
		printf("Content-Type: text/html\n\n");
		printf("<BODY bgcolor=\"#eeeebb\">\n");
		printf("<FONT face=\"fixed,terminal\"\n");
		for(i=0;i<yp;i++)
		{
			printf("%s<br>\n", trimify(data[i]));
		}
		lyp=yp;
		redraw=0;
	}
	else
	{
		for(i=lyp; i<yp; i++)
			printf("%s<br>\n", trimify(data[i]));
		lyp=yp;
	}
}


static void wait_message(int fd, int cmdfd)
{
	fd_set in;
	int n=fd;
	
	FD_ZERO(&in);
	FD_SET(1, &in);
	FD_SET(fd, &in);
	FD_SET(cmdfd, &in);

	if(cmdfd>fd)
		n=cmdfd;
		
	select(n+1, &in, NULL, NULL, NULL);
	
	if(FD_ISSET(cmdfd, &in))
	{
		int tfd=accept(cmdfd, NULL, 0);
		int l;
		char buf[256];
		if(tfd==-1)
			return;
		l=read(tfd, buf, 255);
		buf[l]=0;
		parseinput(buf);
		close(tfd);
	}
}


static char *clean(char *p)
{
	static char buf[256];
	char *x=buf;
	int n=0;
	
	while(*p && n<255)
	{
		if(!isalnum(*p) && *p!='/' && *p!='%' && *p!='?' && *p!=':' && *p!='.' && *p!='~')
			p++;
		else
		{
			*x++=*p++;
			n++;
		}
	}
	*x=0;
	return buf;
}

void print_bit(char c)
{
	static char blob[8192];
	static int blobp;

	if(c=='\n')
	{
		blob[blobp]=0;
		if(strncmp(blob,"VIEW-URL:", 9)==0)
		{
			fflush(stdout);
			printf("<SCRIPT>\n");
			printf("top.frames[0].location=\'%s\'\n",
				clean(blob+9));
			printf("</SCRIPT>\n");
			fflush(stdout);
		}
		else 
		{
			print_string(blob);
			print_string("\n");
		}
		blobp=0;
	}
	else if(blobp<8191)
		blob[blobp++]=c;
}
	
	
void print_out(char *p)
{
	while(*p)
	{
		print_bit(*p);
		p++;
	}
}

static void read_message(int fd)
{
	int l=server_input();	
	if(l==0)
	{
		print_string("\n**Connection Closed!\n\n");
		printf("<SCRIPT>\n");
		printf("alert('Connection has closed')\n");
		printf("setTimeout('self.parent.close()', 4000)\n");
		printf("</SCRIPT>\n");
		weboid_sync();
		exit(0);
	}
}


static int make_unix_socket(void)
{
	struct sockaddr_un sun;
	int fd;
	char name[256];

	snprintf(name, 	256, WEB_TTY_PATH, self.uname, self.flags[1]=getpid());
	unlink(name);
	
	sun.sun_family=AF_UNIX;
	strcpy(sun.sun_path, name);
	
	fd=socket(PF_UNIX, SOCK_STREAM, 0);
	if(fd==-1)
	{
		perror("socket");
		exit(1);
	}
	
	if(bind(fd, (struct sockaddr *)&sun, sizeof(sun))!=0)
	{
		perror("bind");
		exit(1);
	}
	
	listen(fd,2);
	
	return fd;
}

		
int main(int argc, char *argv[])
{
	int fd;
	int cmdfd;
	char *p;

	signal(SIGPIPE, SIG_DFL);
	
	printf("HTTP/1.0 200 OK\n");
	printf("Content-Type: multipart/x-mixed-replace;boundary=--magicmetamarkerthing--\n");
//	printf("--magicmetamarkerthing--\n");

	cgi_process();
	if(load_secret()<0)
	{
		printf("\nNo/bad secret.\n");
		exit(1);
	}
	
	p=getenv("HTTP_COOKIE");
	if(p==NULL)
	{
		init_portaloo(&self, "unset");
	}
	else
	{
		switch(decode_portaloo_cookie(&self, p))
		{
			case -1:
			case 0:
				init_portaloo(&self, "broken");
				break;
			default:
				if(strcmp(self.uname, "broken")==0)
					strcpy(self.uname, "unset-ok");
		}
	}

	if(!(self.flags[0]&FLAGS0_NAME_SET))
	{
		/* Do an error here */
		/* FIXME */
		printf("\nNo login no play.\n");
		exit(1);
	}
		

	memset(data, ' ', LINE_ROWS*LINE_COLS);

		
	cmdfd=make_unix_socket();

	issue_portaloo_cookie(&self);
	print_string("\n\nConnecting as ");
	print_string(self.uname);
	print_string("... \n");

	weboid_sync();
	printf("\n<SCRIPT>\n");
	printf("top.frames[2].location=\'"WEB_TTY_CGI"\'\n");
	printf("</SCRIPT>\n");

	weboid_sync();
	fflush(stdout);
	
	fd=host_connect("linux.mit.edu", 6667);
	
	weboid_sync();
	fflush(stdout);
	irc_setup(fd);
	
	while(1)
	{
		wait_message(fd, cmdfd);
		read_message(fd);
		weboid_sync();
		fflush(stdout);
	}
}
	