#define COMMANDCHAR	'/'
/* each line of history adds 512 bytes to resident size */
#define HISTLEN		8
#define RELEASE		"TinyIRC pre1.0"
/* tinyirc 1.0
   THIS VERSION OF TINYIRC WORKS ONLY WITH 2.8.x IRC SERVERS!!!!

   TinyIRC Alpha Release
   Copyright (C) 1994 Nathan I. Laredo

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License Version 1
   as published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Send your comments and all your spare pocket change to
   laredo@gnu.ai.mit.edu (Nathan Laredo) or to 1604 Lilac Lane,
   Plano, TX 75074, USA.  Any donations are welcome.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <pwd.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>

#include "web_tty.h"
#include "portaloo.h"

struct dlist {
	char name[64];
	struct dlist *next;
};

static struct dlist *object = NULL, *objlist = NULL, *newobj;
static int sockfd, histline;
static char *linein, lineout[512], *history[HISTLEN], localhost[64], *tok_in[256], *tok_out[256], *tmp, serverdata[512], *fromhost, IRCNAME[10], inputbuf[512];
static int cursd = 0;

static time_t idletimer,wasdate;

static struct dlist *additem(item, ptr)
char *item;
struct dlist *ptr;
{
	newobj = (struct dlist *) malloc(sizeof(struct dlist));
	strcpy(newobj->name, item);
	newobj->next = ptr;
	return newobj;
}

static struct dlist *finditem(item, ptr)
char *item;
struct dlist *ptr;
{
	while (ptr != NULL)
		if (strcasecmp(item, ptr->name) == 0)
			break;
		else
			ptr = ptr->next;
	return ptr;
}

static struct dlist *removeitem(item, ptr)
char *item;
struct dlist *ptr;
{
	struct dlist *prev = NULL, *start = ptr;
	while (ptr != NULL)
		if (strcasecmp(item, ptr->name) == 0) {
			newobj = ptr->next;
			if (object == ptr)
				object = NULL;
			free(ptr);
			if (prev == NULL)
				return newobj;
			else {
				prev->next = newobj;
				return start;
			}
		} else {
			prev = ptr;
			ptr = ptr->next;
		}
	return start;
}

static int sendline(void)
{
	if (write(sockfd, lineout, strlen(lineout)) < 1)
		return 0;
	return 1;
}

static int nop(void)
{
	return 1;
}

static int doerror(void)
{
	print_out("*** ERROR:");
	return 2;
}

static int doinvite(void)
{
	char buf[256];
	snprintf(buf, 256, "*** %s (%s) invites you to join %s.",
	       tok_in[0], fromhost, &tok_in[2][1]);
	print_out(buf);
	return 0;
}

static int dojoin(void)
{
	if (strcmp(tok_in[0], IRCNAME) == 0) {
		object = objlist = additem(tok_in[2], objlist);
		wasdate = 0;
		print_out("*** Now talking in ");
		print_out(object->name);
	}
	else
	{
		char buf[256];
		snprintf(buf,256, "*** %s (%s) joined %s", tok_in[0], fromhost, tok_in[2]);
		print_out(buf);
	}
	return 0;
}

static int dokick(void)
{
	char buf[256];
	snprintf(buf,256,"*** %s was kicked from %s by %s (%s)",
	       tok_in[3], tok_in[2], tok_in[0], tok_in[4]);
	print_out(buf);
	if (strcmp(tok_in[3], IRCNAME) == 0) {
		objlist = removeitem(tok_in[2], objlist);
		if (object == NULL)
			object = objlist;
		if (object != NULL)
		{
			snprintf(buf, 256, "\n\r*** Now talking in %s", object->name);
			print_out(buf);
		}
		wasdate = 0;
	}
	return 0;
}

static int dokill(void)
{
	char buf[256];
	snprintf(buf,256, "*** %s killed by %s (%s)", tok_in[3], tok_in[0], tok_in[4]);
	print_out(buf);
	return 0;
}

static int domode(void)
{
	char buf[256];
	snprintf(buf,256, "*** %s changed %s to:", tok_in[0], tok_in[2]);
	print_out(buf);
	return 3;
}

static int donick(void)
{
	char buf[256];
	if (strcmp(tok_in[0], IRCNAME) == 0) {
		wasdate = 0;
		strcpy(IRCNAME, tok_in[2]);
	}
	snprintf(buf,256,"*** %s is now known as %s", tok_in[0], tok_in[2]);
	print_out(buf);
	return 0;
}

static int donotice(void)
{
	char buf[256];
	if (*tok_in[2] != '#')
		snprintf(buf, 256, "-%s-", tok_in[0]);
	else
		snprintf(buf, 256, "-%s:%s-", tok_in[0], tok_in[2]);
	print_out(buf);
	return 3;
}

static int dopart(void)
{
	char buf[256];
	snprintf(buf, 256, "*** %s (%s) left %s", tok_in[0], fromhost,
	       tok_in[2]);
	print_out(buf);
	if (strcmp(tok_in[0], IRCNAME) == 0) {
		objlist = removeitem(tok_in[2], objlist);
		if (object == NULL)
			object = objlist;
		if (object != NULL)
		{
			snprintf(buf, 256, "\n\r*** Now talking in %s", object->name);
			print_out(buf);
		}
		wasdate = 0;
	}
	return 0;
}

static int dopong(void)
{
	print_out("*** Got PONG from ");
	print_out(tok_in[0]);
	print_bit(':');
	return 3;
}

static int doprivmsg(void)
{
	int i = 4, noctcp = 0;
	if (*(tok_in[3]) == '\01') {
		if (*(tok_in[3] + 1) == 'A') {
			print_out("* ");
			print_out(tok_in[0]);
			tok_in[3][strlen(tok_in[3])-1]=0;
			print_bit(' ');
			print_out(tok_in[3]+7);
			noctcp = 1;
			i=4;
		}
		else
		{
			char buf[256];
			tok_in[3][strlen(tok_in[3])-1]=0;
			snprintf(buf, 256, "*** got CTCP %s from %s", tok_in[3]+1, tok_in[0]);
			print_out(buf);
		}
	}
	else
	{
		if (*tok_in[2] != '#')
		{
			print_bit('*');
			print_out(tok_in[0]);
			print_bit('*');
			i = 3;
			noctcp = 1;
		}
		else if (object != NULL && strcasecmp(object->name, tok_in[2]))
		{
			print_bit('<');
			print_out(tok_in[0]);
			print_bit(':');
			print_out(tok_in[2]);
			print_bit('>');
		}
		else
		{
			print_bit('<');
			print_out(tok_in[0]);
			print_bit('>');
		}
		print_bit(' ');
		i = 3;
		noctcp = 1;
	}
	if (noctcp)
		return i;
/* CTCP requests */
	if (*tok_in[3] == 'V') {
		snprintf(lineout, sizeof(lineout), "NOTICE %s \01VERSION TinyIRC 0.4 *IX :0\01\n",
			tok_in[0]);
		sendline();
	} else if (*tok_in[3] == 'P') {
		snprintf(lineout, sizeof(lineout), "NOTICE %s \01PID %d\01\n", tok_in[0], getpid());
		sendline();
	} else if (*tok_in[3] == 'F') {
		snprintf(lineout, sizeof(lineout), "NOTICE %s \01FINGER %s (%s@%s[Portaloo User]) Idle %ld seconds\01\n",
			tok_in[0], self.uname, self.uname, localhost,
			time(NULL) - idletimer);
		sendline();
	} else if (*tok_in[3] == 'C') {
		snprintf(lineout, sizeof(lineout), "NOTICE %s \01CLIENTINFO %s\01\n",
			tok_in[0], "VERSION FINGER CLIENTINFO PID ERRMSG");
		sendline();
	} else if (*tok_in[3] == 'E') {
		snprintf(lineout, sizeof(lineout), "NOTICE %s \01%s %s %s %s\n", tok_in[0],
			tok_in[3], tok_in[4], tok_in[5], tok_in[6]);
		print_out(lineout);
	} else if (*tok_in[3] == 'D') {
		snprintf(lineout, sizeof(lineout), "NOTICE %s \01DCC not supported\01\n",
			tok_in[0]);
		sendline();
	} else {
		snprintf(lineout, sizeof(lineout), "NOTICE %s \01ERRMSG %s\01\n", tok_in[0],
			"I'm sorry dave, I'm afraid I can't do that");
		sendline();
	}
	return 0;
}

static int doquit(void)
{
	char buf[256];
	snprintf(buf, 256,"*** %s (%s) Quit (%s)", tok_in[0], fromhost, tok_in[2]);
	print_out(buf);
	return 0;
}

static int dosquit(void)
{
	return 1;
}

static int dotime(void)
{
	return 1;
}

static int dotopic(void)
{
	char buf[256];
	snprintf(buf, 256, "*** %s set %s topic to \"%s\"", tok_in[0], tok_in[2],
	       tok_in[3]);
	print_out(buf);
	return 0;
}

int donumeric(int num)
{
	char buf[256];
	if (num == 352) {	/* rpl_whoreply */
		snprintf(buf, 256, "%-14s %-10s %-3s %s@%s :", tok_in[3], tok_in[7],
				tok_in[8], tok_in[4], tok_in[5]);
		print_out(buf);
		return 9;
	} else if (num == 432 || num == 433) {
		snprintf(buf, 256, "*** You've chosen an invalid nick.  Choose again.");
		print_out(buf);
		printf("<SCRIPT>\n");
		printf("alert('Connection has closed')\n");
		printf("setTimeout('self.parent.close()', 4000)\n");
		printf("</SCRIPT>\n");
		weboid_sync();
		weboid_sync();
		exit(1);
	} else {		/* all remaining numerics */
//		print_out(tok_in[1]);
		return 3;
	}
}
/* list of all ircd 2.8.16 server commands - alpha order for searching */
#define LISTSIZE 49
static char *cmdlist[LISTSIZE] =
{"ADMIN", "AWAY", "CLOSE", "CONNECT", "DIE", "DNS", "ERROR", "HASH",
"HELP", "INFO", "INVITE", "ISON", "JOIN", "KICK", "KILL", "LINKS", "LIST",
 "LUSERS", "MODE", "MOTD", "MSG", "NAMES", "NICK", "NOTE", "NOTICE", "OPER",
 "PART", "PASS", "PING", "PONG", "PRIVMSG", "QUIT", "REHASH", "RESTART",
 "SERVER", "SQUIT", "STATS", "SUMMON", "TIME", "TOPIC", "TRACE", "USER",
 "USERHOST", "USERS", "VERSION", "WALLOPS", "WHO", "WHOIS", "WHOWAS"};
#define DO_JOIN 12
#define DO_MSG 20
#define DO_PRIVMSG 30
static int numargs[LISTSIZE] =
{1, 1, 1, 3, 1, 1, 1, 1, 1, 1, 2, 1, 1, 3, 2, 1, 1, 2, 2, 1, 2, 1, 1, 1,
 2, 2, 1, 1, 1, 1, 2, 1, 1, 1, 3, 2, 1, 3, 1, 2, 2, 4, 1, 1, 1, 1, 1, 1,
 1};
static int (*docommand[LISTSIZE]) () =
{
	nop, nop, nop, nop, nop, nop, doerror, nop, nop, nop, doinvite, nop,
	    dojoin, dokick, dokill, nop, nop, nop, domode, nop, nop, nop, donick,
	    nop, donotice, nop, dopart, nop, nop, dopong, doprivmsg, doquit, nop,
	    nop, nop, dosquit, nop, nop, dotime, dotopic, nop, nop, nop, nop, nop,
	    nop, nop, nop, nop
};

static int parsedata(void)
{
	int i, found = 0;

/* take care of ping silently */
	if (serverdata[0] == 'P') {
		sprintf(lineout, "PONG :%s\n", IRCNAME);
		return sendline();
	}

//	print_out(">>");
//	print_out(serverdata);
//	print_out(">>\n");
	weboid_sync();
	tok_in[i = 0] = serverdata;
	tok_in[i]++;
	while (tok_in[i] != NULL)
		if (*tok_in[i] == ':')
			break;
		else {
			if ((tmp = strchr(tok_in[i], ' '))!=NULL) {
				tok_in[++i] = &tmp[1];
				*tmp = '\0';
			} else
				tok_in[++i] = NULL;
		}
	if (tok_in[i] != NULL)
		tok_in[i]++;
	tok_in[++i] = NULL;

	if ((tmp = strchr(tok_in[0], '!'))!=NULL) {
		fromhost = &tmp[1];
		*tmp = '\0';
	} else
		fromhost = NULL;

	if ((i = atoi(tok_in[1]))!=0)
		i = donumeric(i);
	else {
		for (i = 0; i < LISTSIZE && !found; i++)
			found = (strcmp(cmdlist[i], tok_in[1]) == 0);
		if (found)
			i = (*docommand[i - 1]) ();
		else
			i = nop();
	}

	if (i)
	{
		while (tok_in[i])	/* output remainder with word wrap */
		{
			print_out(tok_in[i++]);
			print_bit(' ');
		}
	}
	print_bit('\n');
	if (strncmp(tok_in[1], "Closing", 7) == 0)
		return 0;
	return 1;
}

int server_input(void)
{
	char ich;
	int v;
	while ((v=read(sockfd, &ich, 1)) > 0)
		if (ich != '\n' && ich != '\r')
			serverdata[cursd++] = ich;
		else {
			serverdata[cursd] = 0;
			if (cursd) {
				cursd = 0;
				return parsedata();
			}
		}
		
	if(v==-1 && errno != EWOULDBLOCK)
		return 0;
	return 1;
}

void parseinput(char *buf)
{
	int i, j, found = 0;
	
	if(buf[strlen(buf)-1]=='\n')
		buf[strlen(buf)-1]=0;
	if (*buf == '\0')
		return;

	linein = buf;
	strcpy(inputbuf, buf);
	tok_out[i = 0] = strtok(inputbuf, " ");
	while((tok_out[++i] = strtok(NULL, " "))!=NULL);
	if (*tok_out[0] == COMMANDCHAR) {
		tok_out[0]++;
		for (i = 0; i < strlen(tok_out[0]) && isalpha(tok_out[0][i]); i++)
			tok_out[0][i] = toupper(tok_out[0][i]);
		for (i = 0; i < LISTSIZE && !found; i++)
			found = (strncmp(cmdlist[i], tok_out[0], strlen(tok_out[0])) == 0);
		i--;
		if (!found) {
			print_out("*** Invalid command ");
			print_out(tok_out[0]);
			print_bit('\n');
			return;
		}
		if (i == DO_JOIN)
			if ((newobj = finditem(tok_out[1], objlist)) != NULL) {
				object = newobj;
				wasdate = 0;
				print_out("*** Now talking in ");
				print_out(object->name);
				print_bit('\n');
				return;
			}
		if (i == DO_MSG)
			i = DO_PRIVMSG;
		strcpy(lineout, cmdlist[i]);
		j = 0;
		print_out("= ");
		print_out(lineout);
		while (tok_out[++j]) {
			sprintf(&lineout[strlen(lineout)], "%s%s", ((j == numargs[i] &&
								     tok_out[j + 1] != NULL) ? " :" : " "), tok_out[j]);
			print_bit(' ');
			print_out(tok_out[j]);
		}
		print_bit('\n');
		strcat(lineout, "\n");
	} else {
		if (object == NULL) {
			print_out("*** Nowhere to send this.\n");
			return;
		}
		snprintf(lineout, sizeof(lineout), "PRIVMSG %s :%s\n", object->name, linein);
		print_out("> ");
		print_out(tok_out[j = 0]);
		while (tok_out[++j])
		{
			print_bit(' ');
			print_out(tok_out[j]);
		}
		print_bit('\n');
	}
	sendline();
	idletimer = time(NULL);
}

void irc_setup(int fd)
{
	char hostname[64];
	int i;
	
	strncpy(IRCNAME, self.uname, sizeof(IRCNAME));

	sockfd=fd;
	
	sprintf(lineout, "NICK :%s\n", IRCNAME);
	sendline();
	gethostname(localhost, 64);
	gethostname(hostname, 64);
	sprintf(lineout, "USER %s %s %s :%s(portaloo user)\n", self.uname,
		localhost, hostname, self.uname);
	sendline();
	idletimer = time(NULL);
	for (i = 0; i < HISTLEN; i++)
		history[i] = (char *) calloc(512, sizeof(char));
	linein = history[histline = 0];
}

